from os import PathLike
from pathlib import Path
from typing import Literal
from typing import Iterable, Generator
from PIL.Image import Image
from fitz import Document
from doc_page_extractor import (
  clip,
  OCRFragment,
  DocExtractor,
  ExtractedResult,
  Layout,
  LayoutClass,
  BaseLayout,
  PlainLayout,
  TableLayout,
  FormulaLayout,
  TableLayoutParsedFormat,
)

from .document import DocumentExtractor, DocumentParams
from .protocol import DocExtractorProtocol
from .utils import contains_cjka
from .types import (
  Block,
  OCRLevel,
  Text,
  TextBlock,
  TextKind,
  FigureBlock,
  TableBlock,
  FormulaBlock,
  TableFormat,
  ExtractedTableFormat,
  PDFPageExtractorProgressReport,
)


def create_pdf_page_extractor(
      device: Literal["cpu", "cuda"],
      model_dir_path: PathLike,
      ocr_level: OCRLevel = OCRLevel.Once,
      extract_formula: bool = True,
      extract_table_format: ExtractedTableFormat | None = None,
      debug_dir_path: PathLike | None = None,
    ) -> "PDFPageExtractor":

  return PDFPageExtractor(
    device=device,
    ocr_level=ocr_level,
    extract_formula=extract_formula,
    extract_table_format=extract_table_format,
    debug_dir_path=debug_dir_path,
    doc_extractor=DocExtractor(
      model_cache_dir=Path(model_dir_path),
      device=device,
    ),
  )

class PDFPageExtractor:
  def __init__(
        self,
        device: Literal["cpu", "cuda"],
        doc_extractor: DocExtractorProtocol,
        ocr_level: OCRLevel = OCRLevel.Once,
        extract_formula: bool = True,
        extract_table_format: ExtractedTableFormat | None = None,
        debug_dir_path: PathLike | None = None,
      ) -> None:

    if device not in ("cpu", "cuda"):
      raise ValueError("Device must be \"cpu\" or \"cuda\".")

    if extract_table_format is None:
      if device == "cpu":
        extract_table_format = ExtractedTableFormat.DISABLE
      else:
        extract_table_format = ExtractedTableFormat.LATEX

    to_pass_table_format: TableLayoutParsedFormat | None = None
    if extract_table_format == ExtractedTableFormat.LATEX:
      to_pass_table_format = TableLayoutParsedFormat.LATEX
    elif extract_table_format == ExtractedTableFormat.MARKDOWN:
      to_pass_table_format = TableLayoutParsedFormat.MARKDOWN
    elif extract_table_format == ExtractedTableFormat.HTML:
      to_pass_table_format = TableLayoutParsedFormat.HTML

    self._doc_extractor: DocumentExtractor = DocumentExtractor(
      doc_extractor=doc_extractor,
      ocr_level=ocr_level,
      extract_formula=extract_formula,
      extract_table_format=to_pass_table_format,
      debug_dir_path=Path(debug_dir_path) if debug_dir_path is not None else None,
    )

  def prepare_models(self) -> None:
    self._doc_extractor.prepare_models()

  def extract(self, pdf: str | Document, report_progress: PDFPageExtractorProgressReport | None = None) -> Generator[Block, None, None]:
    for _, blocks, _ in self.extract_enumerated_blocks_and_image(
      pdf=pdf,
      report_progress=report_progress,
    ):
      yield from blocks

  def extract_enumerated_blocks_and_image(
      self,
      pdf: str | Document,
      page_indexes: Iterable[int] | None = None,
      report_progress: PDFPageExtractorProgressReport | None = None,
    ) -> Generator[tuple[int, list[Block], Image], None, None]:

    for page_index, result, layouts in self._doc_extractor.extract(DocumentParams(
      pdf=pdf,
      page_indexes=page_indexes,
      report_progress=report_progress,
    )):
      blocks = self._convert_to_blocks(result, layouts)
      page_range = self._texts_range(blocks)

      for block in blocks:
        if not isinstance(block, TextBlock) or \
           block.kind == TextKind.ABANDON:
          continue

        if len(block.texts) == 1:
          mean_line_height, x1, x2 = page_range
        else:
          mean_line_height, x1, x2 = self._texts_range((block,))

        first_text = block.texts[0]
        last_text = block.texts[-1]
        first_delta_x = (first_text.rect.lt[0] + first_text.rect.lb[0]) / 2 - x1
        last_delta_x = x2 - (last_text.rect.rt[0] + last_text.rect.rb[0]) / 2
        block.has_paragraph_indentation = first_delta_x > mean_line_height
        block.last_line_touch_end = last_delta_x < mean_line_height

      yield page_index, blocks, result.extracted_image

  def _convert_to_blocks(self, result: ExtractedResult, layouts: list[Layout]) -> list[Block]:
    store: list[tuple[Layout, Block]] = []
    for layout in layouts:
      if isinstance(layout, PlainLayout):
        self._fill_plain_layout(store, layout, result)
      elif isinstance(layout, TableLayout):
        store.append((layout, self._transform_table(layout, result)))
      elif isinstance(layout, FormulaLayout):
        store.append((layout, self._transform_formula(layout, result)))

    self._fill_font_size_for_blocks(store)
    return [block for _, block in store]

  def _texts_range(self, blocks: Iterable[Block]) -> tuple[float, float, float]:
    sum_lines_height: float = 0.0
    texts_count: int = 0
    x1: float = float("inf")
    x2: float = float("-inf")

    for block in blocks:
      if not isinstance(block, TextBlock):
        continue
      if block.kind == TextKind.ABANDON:
        continue
      for text in block.texts:
        sum_lines_height += text.rect.size[1]
        texts_count += 1
        for x, _ in text.rect:
          x1 = min(x1, x)
          x2 = max(x2, x)

    if texts_count == 0:
      return 0.0, 0.0, 0.0
    return sum_lines_height / texts_count, x1, x2

  def _fill_plain_layout(
        self,
        store: list[tuple[Layout, Block]],
        layout: PlainLayout,
        result: ExtractedResult,
      ):

    def previous_block(cls: LayoutClass) -> Block | None:
      nonlocal store
      for i in range(len(store) - 1, -1, -1):
        layout, block = store[i]
        if cls == layout.cls:
          return block
        if cls != LayoutClass.ABANDON:
          return None
      return None

    cls = layout.cls
    if cls == LayoutClass.TITLE:
      store.append((layout, TextBlock(
        rect=layout.rect,
        kind=TextKind.TITLE,
        font_size=0.0,
        texts=self._convert_to_text(layout.fragments),
      )))
    elif cls == LayoutClass.PLAIN_TEXT:
      store.append((layout, TextBlock(
        rect=layout.rect,
        kind=TextKind.PLAIN_TEXT,
        font_size=0.0,
        texts=self._convert_to_text(layout.fragments),
      )))
    elif cls == LayoutClass.ABANDON:
      store.append((layout, TextBlock(
        rect=layout.rect,
        kind=TextKind.ABANDON,
        font_size=0.0,
        texts=self._convert_to_text(layout.fragments),
      )))
    elif cls == LayoutClass.FIGURE:
      store.append((layout, FigureBlock(
        rect=layout.rect,
        texts=[],
        font_size=0.0,
        image=clip(result, layout),
      )))
    elif cls == LayoutClass.FIGURE_CAPTION:
      block = previous_block(LayoutClass.FIGURE)
      if block is not None:
        assert isinstance(block, FigureBlock)
        block.texts.extend(self._convert_to_text(layout.fragments))
    elif cls == LayoutClass.TABLE_CAPTION or \
          cls == LayoutClass.TABLE_FOOTNOTE:
      block = previous_block(LayoutClass.TABLE)
      if block is not None:
        assert isinstance(block, TableBlock)
        block.texts.extend(self._convert_to_text(layout.fragments))
    elif cls == LayoutClass.FORMULA_CAPTION:
      block = previous_block(LayoutClass.ISOLATE_FORMULA)
      if block is not None:
        assert isinstance(block, FormulaBlock)
        block.texts.extend(self._convert_to_text(layout.fragments))

  def _transform_table(self, layout: TableLayout, result: ExtractedResult) -> TableBlock:
    parsed = layout.parsed
    format: TableFormat = TableFormat.UNRECOGNIZABLE
    content: str = ""

    if parsed is not None and self._can_use_latex(layout):
      content, layout_format = parsed
      if layout_format == TableLayoutParsedFormat.LATEX:
        format = TableFormat.LATEX
      elif layout_format == TableLayoutParsedFormat.MARKDOWN:
        format = TableFormat.MARKDOWN
      elif layout_format == TableLayoutParsedFormat.HTML:
        format = TableFormat.HTML

    return TableBlock(
      rect=layout.rect,
      texts=[],
      font_size=0.0,
      format=format,
      content=content,
      image=clip(result, layout),
    )

  def _transform_formula(self, layout: FormulaLayout, result: ExtractedResult) -> FormulaBlock:
    content: str | None = None
    if layout.latex is not None and self._can_use_latex(layout):
      content = layout.latex

    return FormulaBlock(
      rect=layout.rect,
      texts=[],
      font_size=0.0,
      content=content,
      image=clip(result, layout),
    )

  def _fill_font_size_for_blocks(self, store: list[tuple[Layout, Block]]):
    font_sizes: list[float] = []

    for layout, _ in store:
      if len(layout.fragments) == 0:
        font_sizes.append(0.0)
      else:
        sum_height: float = 0.0
        for fragment in layout.fragments:
          sum_height += fragment.rect.size[1]

        # without considering vertical writing, the height of a line of text is proportional to the font size.
        font_size = sum_height / len(layout.fragments)
        font_sizes.append(font_size)

    max_font_size: float = 1.0
    min_font_size: float = 1.0

    if len(font_sizes) > 0:
      max_font_size = max(font_sizes)
      min_font_size = min(font_sizes)

    if max_font_size == min_font_size:
      for _, block in store:
        block.font_size = 0
    else:
      for font_size, (_, block) in zip(font_sizes, store):
        block.font_size = (font_size - min_font_size) / (max_font_size - min_font_size)

  def _convert_to_text(self, fragments: list[OCRFragment]) -> list[Text]:
    return [
      Text(
        content=f.text,
        rank=f.rank,
        rect=f.rect,
      )
      for f in fragments
    ]

  def _can_use_latex(self, layout: BaseLayout) -> bool:
    return all(
      not contains_cjka(fragment.text)
      for fragment in layout.fragments
    )


import json

from os import PathLike
from pathlib import Path
from typing import Literal
from uuid import uuid4
from zipfile import ZipFile
from xml.etree.ElementTree import fromstring, Element
from .types import TableRender, LaTeXRender
from .gen_part import generate_part
from .gen_index import gen_index, NavPoint
from .i18n import I18N
from .template import Template
from .context import Context


def generate_epub_file(
      from_dir_path: PathLike,
      epub_file_path: PathLike,
      lan: Literal["zh", "en"] = "zh",
      table_render: TableRender = TableRender.HTML,
      latex_render: LaTeXRender = LaTeXRender.MATHML,
    ) -> None:

  i18n = I18N(lan)
  template = Template()
  from_dir_path = Path(from_dir_path)
  epub_file_path = Path(epub_file_path)
  index_path = from_dir_path / "index.json"
  meta_path = from_dir_path / "meta.json"
  assets_path: Path | None = from_dir_path / "assets"
  chapters_path: Path = from_dir_path / "chapters"
  head_chapter_path = chapters_path / "chapter.xml"

  toc_ncx: str
  nav_points: list[NavPoint] = []
  meta: dict = {}
  has_head_chapter: bool = head_chapter_path.exists()
  has_cover: bool = (from_dir_path / "cover.png").exists()

  if meta_path.exists():
    with open(meta_path, "r", encoding="utf-8") as f:
      meta = json.loads(f.read())

  if not assets_path.exists():
    assets_path = None

  toc_ncx, nav_points = gen_index(
    template=template,
    i18n=i18n,
    meta=meta,
    index_file_path=index_path,
    has_cover=has_cover,
    check_chapter_exits=lambda id: (chapters_path / f"chapter_{id}.xml").exists(),
  )
  epub_base_path = epub_file_path.parent
  epub_base_path.mkdir(parents=True, exist_ok=True)

  with ZipFile(epub_file_path, "w") as file:
    context = Context(
      file=file,
      assets_path=assets_path,
      table_render=table_render,
      latex_render=latex_render,
    )
    file.writestr(
      zinfo_or_arcname="mimetype",
      data=template.render("mimetype").encode("utf-8"),
    )
    file.writestr(
      zinfo_or_arcname="OEBPS/toc.ncx",
      data=toc_ncx.encode("utf-8"),
    )
    _write_chapters(
      context=context,
      template=template,
      i18n=i18n,
      nav_points=nav_points,
      chapters_path=chapters_path,
      has_head_chapter=has_head_chapter,
      head_chapter_path=head_chapter_path,
    )
    _write_basic_files(
      context=context,
      template=template,
      i18n=i18n,
      meta=meta,
      nav_points=nav_points,
      has_cover=has_cover,
      has_head_chapter=has_head_chapter,
    )
    _write_assets(
      context=context,
      template=template,
      i18n=i18n,
      from_dir_path=from_dir_path,
      has_cover=has_cover,
    )

def _write_assets(
    context: Context,
    template: Template,
    i18n: I18N,
    from_dir_path: Path,
    has_cover: bool,
  ):
  context.file.writestr(
    zinfo_or_arcname="OEBPS/styles/style.css",
    data=template.render("style.css").encode("utf-8"),
  )
  if has_cover:
    context.file.writestr(
      zinfo_or_arcname="OEBPS/Text/cover.xhtml",
      data=template.render(
        template="cover.xhtml",
        i18n=i18n,
      ).encode("utf-8"),
    )
  if has_cover:
    context.file.write(
      filename=from_dir_path / "cover.png",
      arcname="OEBPS/assets/cover.png",
    )
  context.add_used_asset_files()

def _write_chapters(
    context: Context,
    template: Template,
    i18n: I18N,
    nav_points: list[NavPoint],
    chapters_path: Path,
    has_head_chapter: bool,
    head_chapter_path: Path,
  ):

  if has_head_chapter:
    chapter_xml = _read_xml(head_chapter_path)
    data = generate_part(context, template, chapter_xml, i18n)
    context.file.writestr(
      zinfo_or_arcname="OEBPS/Text/head.xhtml",
      data=data.encode("utf-8"),
    )
  for nav_point in nav_points:
    chapter_path = chapters_path / f"chapter_{nav_point.index_id}.xml"
    if chapter_path.exists():
      chapter_xml = _read_xml(chapter_path)
      data = generate_part(context, template, chapter_xml, i18n)
      context.file.writestr(
        zinfo_or_arcname="OEBPS/Text/" + nav_point.file_name,
        data=data.encode("utf-8"),
      )

def _write_basic_files(
    context: Context,
    template: Template,
    i18n: I18N,
    meta: dict,
    nav_points: list[NavPoint],
    has_cover: bool,
    has_head_chapter: bool,
  ):
  context.file.writestr(
    zinfo_or_arcname="META-INF/container.xml",
    data=template.render("container.xml").encode("utf-8"),
  )
  content = template.render(
    template="content.opf",
    meta=meta,
    i18n=i18n,
    ISBN=meta.get("ISBN", str(uuid4())),
    nav_points=nav_points,
    has_head_chapter=has_head_chapter,
    has_cover=has_cover,
    asset_files=context.used_files,
  )
  context.file.writestr(
    zinfo_or_arcname="OEBPS/content.opf",
    data=content.encode("utf-8"),
  )

def _read_xml(path: Path) -> Element:
  with open(path, "r", encoding="utf-8") as file:
    return fromstring(file.read())


import json
import datetime

from os import PathLike
from pathlib import Path
from typing import cast, Any, Generator, Sequence
from importlib.resources import files
from jinja2 import Environment, Template
from xml.etree.ElementTree import Element
from pydantic import SecretStr
from logging import getLogger, DEBUG, Formatter, Logger, FileHandler
from tiktoken import get_encoding, Encoding
from langchain_core.messages import SystemMessage, HumanMessage

from ..template import create_env
from ..xml import decode_friendly, encode_friendly
from .increasable import Increasable
from .executor import LLMExecutor


class LLM:
  def __init__(
      self,
      key: str,
      url: str,
      model: str,
      token_encoding: str,
      timeout: float | None = None,
      top_p: float | tuple[float, float] | None = None,
      temperature: float | tuple[float, float] | None = None,
      retry_times: int = 5,
      retry_interval_seconds: float = 6.0,
      log_dir_path: PathLike | None = None,
    ):
    prompts_path = files("pdf_craft") / "data" / "prompts"
    self._templates: dict[str, Template] = {}
    self._encoding: Encoding = get_encoding(token_encoding)
    self._env: Environment = create_env(prompts_path)
    self._logger_save_path: Path | None = None

    if log_dir_path is not None:
      self._logger_save_path = Path(log_dir_path)
      if not self._logger_save_path.exists():
        self._logger_save_path.mkdir(parents=True, exist_ok=True)
      elif not self._logger_save_path.is_dir():
        self._logger_save_path = None

    self._executor = LLMExecutor(
      url=url,
      model=model,
      api_key=cast(SecretStr, key),
      timeout=timeout,
      top_p=Increasable(top_p),
      temperature=Increasable(temperature),
      retry_times=retry_times,
      retry_interval_seconds=retry_interval_seconds,
      create_logger=self._create_logger,
    )

  def _create_logger(self) -> Logger | None:
    if self._logger_save_path is None:
      return None

    now = datetime.datetime.now(datetime.timezone.utc)
    timestamp = now.strftime("%Y-%m-%d %H-%M-%S %f")
    file_path = self._logger_save_path / f"request {timestamp}.log"
    logger = getLogger(f"LLM Request {timestamp}")
    logger.setLevel(DEBUG)
    handler = FileHandler(file_path, encoding="utf-8")
    handler.setLevel(DEBUG)
    handler.setFormatter(Formatter("%(asctime)s    %(message)s", "%H:%M:%S"))
    logger.addHandler(handler)

    return logger

  def request_markdown(self, template_name: str, user_data: Element | str, params: dict[str, Any] | None = None) -> str:
    if params is None:
      params = {}
    return self._executor.request(
      input=self._create_input(template_name, user_data, params),
      parser=self._encode_markdown,
    )

  def request_json(self, template_name: str, user_data: Element | str, params: dict[str, Any] | None = None) -> Any:
    if params is None:
      params = {}
    return self._executor.request(
      input=self._create_input(template_name, user_data, params),
      parser=self._encode_json,
    )

  def request_xml(self, template_name: str, user_data: Element | str, params: dict[str, Any] | None = None) -> Element:
    if params is None:
      params = {}
    return self._executor.request(
      input=self._create_input(template_name, user_data, params),
      parser=self._encode_xml,
    )

  def _create_input(self, template_name: str, user_data: Element | str, params: dict[str, Any]):
    data: str
    if isinstance(user_data, Element):
      data = encode_friendly(user_data)
      data = f"```XML\n{data}\n```"
    else:
      data = user_data

    template = self._template(template_name)
    prompt = template.render(**params)
    return [
      SystemMessage(content=prompt),
      HumanMessage(content=data)
    ]

  def prompt_tokens_count(self, template_name: str, params: dict[str, Any]) -> int:
    template = self._template(template_name)
    prompt = template.render(**params)
    return len(self._encoding.encode(prompt))

  def encode_tokens(self, text: str) -> list[int]:
    return self._encoding.encode(text)

  def decode_tokens(self, tokens: Sequence[int]) -> str:
    return self._encoding.decode(tokens)

  def count_tokens_count(self, text: str) -> int:
    return len(self._encoding.encode(text))

  def _template(self, template_name: str) -> Template:
    template = self._templates.get(template_name, None)
    if template is None:
      template = self._env.get_template(template_name)
      self._templates[template_name] = template
    return template

  def _encode_markdown(self, response: str) -> str:
    for quote in self._search_quotes("markdown", response):
      return quote
    raise ValueError("No valid Markdown response found")

  def _encode_json(self, response: str) -> Any:
    for quote in self._search_quotes("json", response):
      return json.loads(quote)
    raise ValueError("No valid Markdown response found")

  def _encode_xml(self, response: str) -> Element:
    for element in decode_friendly(response, "response"):
      return element
    raise ValueError("No valid XML response found")

  def _search_quotes(self, kind: str, response: str) -> Generator[str, None, None]:
    start_marker = f"```{kind}"
    end_marker = "```"
    start_index = 0

    while True:
      start_index = self._find_ignore_case(
        raw=response,
        sub=start_marker,
        start=start_index,
      )
      if start_index == -1:
        break

      end_index = self._find_ignore_case(
        raw=response,
        sub=end_marker,
        start=start_index + len(start_marker),
      )
      if end_index == -1:
        break

      extracted_text = response[start_index + len(start_marker):end_index].strip()
      yield extracted_text
      start_index = end_index + len(end_marker)

  def _find_ignore_case(self, raw: str, sub: str, start: int = 0):
    if not sub:
      return 0 if 0 >= start else -1

    raw_len, sub_len = len(raw), len(sub)
    for i in range(start, raw_len - sub_len + 1):
      match = True
      for j in range(sub_len):
        if raw[i + j].lower() != sub[j].lower():
          match = False
          break
      if match:
        return i
    return -1


from xml.etree.ElementTree import Element
from .tag import Tag, TagKind


def tag_to_element(tag: Tag) -> Element:
  element = Element(tag.name)
  for attr_name, attr_value in tag.attributes:
    element.set(attr_name, attr_value)
  return element

def element_to_tag(element: Element, kind: TagKind, proto: str = "") -> Tag:
  tag = Tag(
    kind=kind,
    name=element.tag,
    proto=proto,
    attributes=[],
  )
  if kind != TagKind.CLOSING:
    for attr_name in sorted(list(element.keys())):
      attr_value = element.get(attr_name, "")
      tag.attributes.append((attr_name, attr_value))

  # To make LLM easier to understand, the naming here is restricted in a more strict way.
  # https://github.com/oomol-lab/pdf-craft/issues/149
  invalid_name = tag.find_invalid_name()
  if invalid_name is not None:
    raise ValueError(f"find invalid tag name or attribute name: {invalid_name}")

  invalid_attr_pair = tag.find_invalid_attr_value()
  if invalid_attr_pair is not None:
    attr_name, attr_value = invalid_attr_pair
    raise ValueError(f"find invalid attribute value: {attr_name}=\"{attr_value}\"")

  return tag


from os import PathLike
from pathlib import Path

from ..llm import LLM
from ..pdf import PDFPageExtractor

from .types import CorrectionMode
from .reporter import Reporter, AnalysingStep, AnalysingStepReport, AnalysingProgressReport
from .window import parse_window_tokens, LLMWindowTokens
from .ocr import generate_ocr_pages
from .sequence import extract_sequences
from .correction import correct, Level as CorrectionLevel
from .meta import extract_meta
from .contents import extract_contents
from .chapter import generate_chapters
from .reference import generate_chapters_with_footnotes
from .output import output
from .utils import MultiThreads


def analyse(
    llm: LLM,
    pdf_page_extractor: PDFPageExtractor,
    pdf_path: PathLike,
    analysing_dir_path: PathLike,
    output_dir_path: PathLike,
    report_step: AnalysingStepReport | None = None,
    report_progress: AnalysingProgressReport | None = None,
    correction_mode: CorrectionMode = CorrectionMode.NO,
    window_tokens: LLMWindowTokens | int | None = None,
    threads_count: int = 1,
  ) -> None:

  if correction_mode == CorrectionMode.DETAILED:
    print(
      "`correction_mode=CorrectionMode.DETAILED` is still experimental and not open to the public yet. " +
      "`correction_mode` will be forced to `CorrectionMode.ONCE`."
    )
    correction_mode = CorrectionMode.ONCE

  window_tokens = parse_window_tokens(window_tokens)
  threads = MultiThreads(threads_count)
  reporter = Reporter(
    report_step=report_step,
    report_progress=report_progress,
  )
  analysing_dir_path = Path(analysing_dir_path)
  ocr_path = analysing_dir_path / "ocr"
  assets_path = analysing_dir_path / "assets"
  sequence_path = analysing_dir_path / "sequence"
  correction_path = analysing_dir_path / "correction"
  contents_path = analysing_dir_path / "contents"
  chapter_path = analysing_dir_path / "chapter"
  reference_path = analysing_dir_path / "reference"

  generate_ocr_pages(
    extractor=pdf_page_extractor,
    reporter=reporter,
    pdf_path=Path(pdf_path),
    ocr_path=ocr_path,
    assets_path=assets_path,
  )
  extract_sequences(
    llm=llm,
    reporter=reporter,
    threads=threads,
    workspace_path=sequence_path,
    ocr_path=ocr_path,
    max_request_data_tokens=window_tokens.max_request_data_tokens,
    max_paragraph_tokens=window_tokens.max_verify_paragraph_tokens,
    max_paragraphs=window_tokens.max_verify_paragraphs_count,
  )
  sequence_output_path = sequence_path / "output"

  if correction_mode != CorrectionMode.NO:
    level: CorrectionLevel
    if correction_mode == CorrectionMode.ONCE:
      level = CorrectionLevel.Single
    elif correction_mode == CorrectionMode.DETAILED:
      level = CorrectionLevel.Multiple
    else:
      raise ValueError(f"Unknown correction mode: {correction_mode}")

    sequence_output_path = correct(
      llm=llm,
      reporter=reporter,
      threads=threads,
      level=level,
      workspace_path=correction_path,
      text_path=sequence_output_path / "text",
      footnote_path=sequence_output_path / "footnote",
      max_data_tokens=window_tokens.max_request_data_tokens,
    )

  reporter.go_to_step(AnalysingStep.EXTRACT_META)
  meta_path = extract_meta(
    llm=llm,
    workspace_path=analysing_dir_path / "meta",
    sequence_path=sequence_output_path / "text",
    max_request_tokens=window_tokens.max_request_data_tokens,
  )
  contents = extract_contents(
    llm=llm,
    reporter=reporter,
    workspace_path=contents_path,
    sequence_path=sequence_output_path / "text",
    max_data_tokens=window_tokens.max_request_data_tokens,
  )
  chapter_output_path, contents = generate_chapters(
    llm=llm,
    reporter=reporter,
    threads=threads,
    contents=contents,
    sequence_path=sequence_output_path / "text",
    workspace_path=chapter_path,
    max_request_tokens=window_tokens.max_request_data_tokens,
  )
  footnote_sequence_path = sequence_output_path / "footnote"

  if footnote_sequence_path.exists():
    chapter_output_path = generate_chapters_with_footnotes(
      reporter=reporter,
      chapter_path=chapter_output_path,
      footnote_sequence_path=footnote_sequence_path,
      workspace_path=reference_path,
    )

  reporter.go_to_step(AnalysingStep.OUTPUT)
  output(
    contents=contents,
    output_path=Path(output_dir_path),
    meta_path=meta_path,
    chapter_output_path=chapter_output_path,
    assets_path=assets_path,
  )



from typing import Generic, TypeVar, Callable
from threading import Lock, Thread


T = TypeVar("T")
P = TypeVar("P")

class MultiThreads:
  def __init__(self, threads_count: int):
    self._threads_count: int = threads_count
    if threads_count <= 0:
      raise ValueError("threads_count must be greater than 0")

  def run(
        self,
        next_task: Callable[[], T | None],
        thread_payload: Callable[[], P],
        invoke: Callable[[P, T], None],
      ) -> None:

    if self._threads_count == 1:
      payload = thread_payload()
      while True:
        task = next_task()
        if task is None:
          break
        invoke(payload, task)
    else:
      invoker: _Invoker[P, T] = _Invoker(
        threads_count=self._threads_count,
        next_task=next_task,
        thread_payload=thread_payload,
        invoke=invoke,
      )
      invoker.do()

class _Invoker(Generic[P, T]):
  def __init__(
        self,
        threads_count: int,
        next_task: Callable[[], T | None],
        thread_payload: Callable[[], P],
        invoke: Callable[[T], None],
      ) -> None:

    self._threads_count: int = threads_count
    self._thread_payload: Callable[[], P] = thread_payload
    self._next_task: Callable[[], T | None] = next_task
    self._invoke: Callable[[T], None] = invoke
    self._threads: list[Thread] = []
    self._task_lock: Lock = Lock()
    self._done: bool = False
    self._error: Exception | None = None

  def do(self):
    for _ in range(self._threads_count):
      payload = self._thread_payload()
      self._threads.append(Thread(
        target=self._run_thread_loop,
        args=(payload,),
      ))
    for thread in self._threads:
      thread.start()
    for thread in self._threads:
      thread.join()
    if self._error is not None:
      raise self._error

  def _run_thread_loop(self, payload: P):
    while True:
      task = self._get_next_task()
      if task is None:
        break
      try:
        self._invoke(payload, task)
      except Exception as error:
        with self._task_lock:
          if self._error is not None:
            break
          self._error = error

  def _get_next_task(self) -> T | None:
    with self._task_lock:
      if self._error is not None:
        return None
      if self._done:
        return None
      task = self._next_task()
      if task is None:
        self._done = True
      return task



from typing import TypedDict
from strenum import StrEnum


class Phase(StrEnum):
  MAPPER = "mapper"
  CHAPTER = "chapter"
  COMPLETED = "completed"

class State(TypedDict):
  phase: Phase
  has_contents: bool
  completed_ranges: list[list[int]]



from xml.etree.ElementTree import Element
from ..data import Paragraph
from ..utils import search_xml_children


def normalize_layout_xml(paragraph: Paragraph) -> Element | None:
  merged_layout: Element | None = None
  for layout in paragraph.layouts:
    if merged_layout is None:
      merged_layout = layout.to_xml()
    else:
      for line in layout.lines:
        merged_layout.append(line.to_xml())

  if merged_layout is not None:
    merged_layout.attrib = {}
    for child, _ in search_xml_children(merged_layout):
      if child.tag == "line":
        child.attrib = {}

  return merged_layout


from typing import TypedDict
from strenum import StrEnum


class Phase(StrEnum):
  INIT = "init"
  COLLECT = "collect"
  ANALYSE = "analyse"
  NO_CONTENTS = "no_contents"
  GENERATED = "generated"

class State(TypedDict):
  phase: Phase
  max_data_tokens: int
  page_indexes: list[int]


import sys

from pathlib import Path
from typing import Generator, TypedDict
from strenum import StrEnum
from abc import ABC, abstractmethod
from xml.etree.ElementTree import Element

from ...llm import LLM
from ...xml import encode_friendly
from ..sequence import read_paragraphs
from ..data import Paragraph, ParagraphType, AssetLayout, FormulaLayout
from ..utils import Context, MultiThreads


class Phase(StrEnum):
  Text = "text"
  FOOTNOTE = "footnote"
  GENERATION = "generation"
  COMPLETED = "completed"

class Level(StrEnum):
  Single = "single"
  Multiple = "multiple"

class State(TypedDict):
  phase: Phase
  level: Level
  max_data_tokens: int
  completed_ranges: list[list[int]]

class Corrector(ABC):
  def __init__(self, llm: LLM, context: Context[State], threads: MultiThreads):
    self.llm: LLM = llm
    self.ctx: Context[State] = context
    self.threads: MultiThreads = threads

  @abstractmethod
  def do(self, from_path: Path, request_path: Path, is_footnote: bool) -> None:
    raise NotImplementedError()

  def extract_lines(self, extracted_element: Element) -> Generator[tuple[tuple[int, int], list[str]], None, None]:
    for layout in extracted_element:
      layout_id = layout.get("id", None)
      if layout_id is None:
        continue
      id1, id2 = layout_id.split("/", maxsplit=1)
      index = (int(id1), int(id2))
      lines: list[str] = []

      for line in layout:
        if line.tag != "line":
          continue
        line_id = line.get("id", None)
        if line_id is None:
          continue
        if line.text:
          lines.append(line.text.strip())
        else:
          lines.append("")

      yield index, lines

  def generate_request_xml(self, from_path: Path) -> Generator[tuple[tuple[int, int], tuple[int, int], Element], None, None]:
    max_data_tokens = self.ctx.state["max_data_tokens"]
    request_element = Element("request")
    request_begin: tuple[int, int] = (sys.maxsize, sys.maxsize)
    request_end: tuple[int, int] = (-1, -1)
    data_tokens: int = 0
    last_type: ParagraphType | None = None

    for paragraph in read_paragraphs(from_path):
      layout_element = self._paragraph_to_layout_xml(paragraph)
      tokens = self.llm.count_tokens_count(
        text=encode_friendly(layout_element),
      )
      if len(request_element) > 0 and (
        data_tokens + tokens > max_data_tokens or
        last_type != paragraph.type
      ):
        yield request_begin, request_end, request_element
        request_element = Element("request")
        data_tokens = 0
        request_begin = (sys.maxsize, sys.maxsize)
        request_end = (-1, -1)

      paragraph_index = (paragraph.page_index, paragraph.order_index)
      request_element.append(layout_element)
      request_begin = min(request_begin, paragraph_index)
      request_end = max(request_end, paragraph_index)
      data_tokens += tokens
      last_type = paragraph.type

    if len(request_element) > 0:
      yield request_begin, request_end, request_element

  def _paragraph_to_layout_xml(self, paragraph: Paragraph) -> tuple[int, Element]:
    layout_element: Element | None = None
    next_line_id: int = 1

    for layout in paragraph.layouts:
      if layout_element is None:
        layout_element = Element(layout.kind.value)
        layout_element.set("id", layout.id)

      if isinstance(layout, AssetLayout):
        line_element = Element("line")
        line_element.set("id", str(object=next_line_id))
        next_line_id += 1
        layout_element.append(line_element)

        if isinstance(layout, FormulaLayout) and layout.latex:
          line_element.text = layout.latex.strip()
        else:
          line_element.text = f"[[here is a {layout.kind.value}]]"

      else:
        for line in layout.lines:
          line_element = Element("line")
          line_element.set("id", str(next_line_id))
          line_element.text = line.text.strip()
          layout_element.append(line_element)
          next_line_id += 1

    assert layout_element is not None
    return layout_element



from typing import Generator
from pathlib import Path
from xml.etree.ElementTree import Element

from ...llm import LLM
from ...xml import encode_friendly
from ..data import Layout, LayoutKind
from ..sequence import read_paragraphs
from ..contents import Contents, Chapter
from ..utils import remove_file, Context, Partition, PartitionTask, MultiThreads
from .common import State
from .fragment import Fragment, FragmentRequest


def map_contents(
      llm: LLM,
      context: Context[State],
      threads: MultiThreads,
      contents: Contents,
      sequence_path: Path,
      map_path: Path,
    ) -> None:

  mapper = _ContentsMapper(
    llm=llm,
    context=context,
    threads=threads,
    contents=contents,
    sequence_path=sequence_path,
    map_path=map_path,
  )
  mapper.do()

_MAX_ABSTRACT_CONTENT_TOKENS = 150

class _ContentsMapper:
  def __init__(
        self,
        llm: LLM,
        context: Context[State],
        threads: MultiThreads,
        contents: Contents,
        sequence_path: Path,
        map_path: Path,
      ) -> None:

    self._llm: LLM = llm
    self._ctx: Context[State] = context
    self._threads: MultiThreads = threads
    self._contents: Contents = contents
    self._sequence_path: Path = sequence_path
    self._map_path: Path = map_path

  def do(self):
    contents_tokens_count = self._llm.count_tokens_count(
      text=encode_friendly(self._get_contents_xml()),
    )
    partition: Partition[tuple[int], State, FragmentRequest] = Partition(
      dimension=1,
      context=self._ctx,
      sequence=self._gen_request(contents_tokens_count),
      done=lambda _, __: self._ctx.reporter.increment(),
      remove=lambda begin, end: remove_file(
        self._map_path / f"pages_{begin}_{end}.xml"
      ),
    )
    with partition:
      self._threads.run(
        next_task=partition.pop_task,
        thread_payload=lambda: None,
        invoke=lambda _, task: self._emit_request(task),
      )

  def _gen_request(self, contents_tokens_count: int) -> Generator[tuple[int, int, FragmentRequest], None, None]:
    request = FragmentRequest()
    request_tokens = 0
    max_request_tokens = self._ctx.state["max_request_tokens"]
    max_request_tokens = max(
      max_request_tokens - contents_tokens_count,
      int(max_request_tokens * 0.25),
    )
    for fragment in self._read_fragment():
      id = 1 # only for calculate tokens. won't be used in request
      request_text = encode_friendly(fragment.to_request_xml(id))
      tokens = len(self._llm.encode_tokens(request_text))
      if request_tokens > 0 and request_tokens + tokens > max_request_tokens:
        yield request.begin_page_index, request.end_page_index, request
        request = FragmentRequest()
        request_tokens = 0

      request_tokens += tokens
      request.append(fragment)

    if request_tokens > 0:
      yield request.begin_page_index, request.end_page_index, request

  # running in background thread
  def _emit_request(self, task: PartitionTask[tuple[int], State, FragmentRequest]):
    with task:
      request = task.payload
      request_xml = request.complete_to_xml()
      request_xml.insert(0, self._get_contents_xml())
      resp_xml = self._llm.request_xml(
        template_name="contents/mapper",
        user_data=request_xml,
        params={
          "fragments_count": request.fragments_count,
        },
      )
      page_indexes_set: set[int] = set()
      map_element = Element("map")
      patch_element = Element("patch")

      for page_index, sub_patch_element in request.generate_patch_xmls(resp_xml):
        page_indexes_set.add(page_index)
        patch_element.append(sub_patch_element)

      for headline_id, chapter_id in request.generate_matched_mapper(resp_xml):
        mapper = Element("mapper")
        map_element.append(mapper)
        mapper.set("headline-id", headline_id)
        mapper.set("chapter-id", str(chapter_id))

      if page_indexes_set:
        page_indexes = sorted(list(page_indexes_set))
        map_element.set("page-indexes", ",".join(map(str, page_indexes)))

      if len(patch_element) > 0:
        map_element.append(patch_element)

      file_name = f"pages_{request.begin_page_index}_{request.end_page_index}.xml"
      file_path = self._map_path / file_name
      self._ctx.write_xml_file(file_path, map_element)

  def _read_fragment(self) -> Generator[Fragment, None, None]:
    fragment: Fragment | None = None
    fragment_tokens_count: int = 0

    for layout in self._read_headline_and_text():
      if layout.kind == LayoutKind.HEADLINE:
        if fragment is not None:
          if fragment.is_abstracts_empty or \
             fragment.page_index != layout.page_index:
            yield fragment
            fragment = None

        if fragment is None:
          fragment = Fragment(layout.page_index)
        fragment.append_headline(layout)

      elif fragment is not None and layout.kind == LayoutKind.TEXT:
        for line in layout.lines:
          tokens = self._llm.encode_tokens(line.text)
          next_tokens_count = fragment_tokens_count + len(tokens)
          if next_tokens_count <= _MAX_ABSTRACT_CONTENT_TOKENS:
            fragment.append_abstract_line(
              parent_layout=layout,
              text=line.text,
            )
            fragment_tokens_count += len(tokens)
          else:
            can_added_tokens_count = next_tokens_count - _MAX_ABSTRACT_CONTENT_TOKENS
            tokens = tokens[:can_added_tokens_count]
            fragment.append_abstract_line(
              parent_layout=layout,
              text=self._llm.decode_tokens(tokens) + "...",
              splitted=True,
            )
            fragment_tokens_count = _MAX_ABSTRACT_CONTENT_TOKENS

          if fragment_tokens_count >= _MAX_ABSTRACT_CONTENT_TOKENS:
            yield fragment
            fragment = None
            fragment_tokens_count = 0
            break

    if fragment is not None:
      yield fragment

  def _read_headline_and_text(self) -> Generator[Layout, None, None]:
    for paragraph in read_paragraphs(self._sequence_path):
      if paragraph.page_index in self._contents.page_indexes:
        continue
      for layout in paragraph.layouts:
        if layout.kind not in (LayoutKind.HEADLINE, LayoutKind.TEXT):
          continue
        if all(self._is_empty(line.text) for line in layout.lines):
          continue
        yield layout

  def _get_contents_xml(self) -> Element:
    contents_element = Element("contents")
    for tag, chapters in (
      ("prefaces", self._contents.prefaces),
      ("chapters", self._contents.chapters),
    ):
      if chapters:
        chapters_element = Element(tag)
        contents_element.append(chapters_element)
        for chapter in chapters:
          chapter_element = self._to_chapter_xml(chapter)
          chapters_element.append(chapter_element)
    return contents_element

  def _to_chapter_xml(self, chapter: Chapter) -> Element:
    chapter_element = Element("chapter")
    chapter_element.set("id", str(chapter.id))
    chapter_element.text = chapter.name
    for child in chapter.children:
      child_xml = self._to_chapter_xml(child)
      chapter_element.append(child_xml)
    return chapter_element

  def _is_empty(self, text: str) -> bool:
    for char in text:
      if char not in (" ", "\n", "\r", "\t"):
        return False
    return True


from dataclasses import dataclass
from pathlib import Path
from typing import Generator, Iterator
from xml.etree.ElementTree import Element

from ..contents import Contents, Chapter
from ..sequence import decode_paragraph
from ..data import Paragraph
from ..utils import xml_files, read_xml_file, XML_Info


def read_paragraphs(paragraph_path: Path) -> Generator[Paragraph, None, None]:
  for file_path, _, page_index, order_index in xml_files(paragraph_path):
    raw_root = read_xml_file(file_path)
    root = Element(raw_root.tag, attrib=raw_root.attrib)
    for layout in raw_root:
      if layout.get("id", None) is None:
        continue
      if _is_invalid_layout_element(layout):
        continue
      root.append(layout)

    yield decode_paragraph(
      element=root,
      page_index=page_index,
      order_index=order_index,
    )

def read_paragraphs_with_patches(
      paragraph_path: Path,
      contents: Contents,
      map_path: Path,
    ) -> Generator[tuple[Chapter | None, Paragraph], None, None]:

  contents_page_indexes = set(contents.page_indexes)
  reader = _PatcherReader(
    map_path=map_path,
    no_matched_chapters=dict((c.id, c) for c in contents),
  )
  for file_path, _, page_index, order_index in xml_files(paragraph_path):
    if page_index in contents_page_indexes:
      continue # skip contents

    raw_root = read_xml_file(file_path)
    root = Element(raw_root.tag, attrib=raw_root.attrib)
    root_chapter: Chapter | None = None

    for raw_layout in raw_root:
      id = raw_layout.get("id", None)
      if id is None:
        continue

      page_index, _ = id.split("/", maxsplit=1)
      page_index = int(page_index)
      patch = reader.read(page_index)

      chapter: Chapter | None = None
      layout: Element = raw_layout
      if patch is not None:
        chapter = patch.headline2chapter.get(id, None)
        layout = patch.layout_xmls_patches.get(id, raw_layout)

      if chapter is not None:
        root_chapter = chapter

      if _is_invalid_layout_element(layout):
        continue # means it was removed

      if root_chapter is not None and len(root) > 0:
        # paragraph maybe splitted by a chapter headline
        yield root_chapter, decode_paragraph(
          element=root,
          page_index=page_index,
          order_index=order_index,
        )
        root = Element(raw_root.tag, attrib=raw_root.attrib)

      root.append(layout)

    if len(root) > 0:
      yield root_chapter, decode_paragraph(
        element=root,
        page_index=page_index,
        order_index=order_index,
      )

def _is_invalid_layout_element(layout: Element) -> bool:
  if len(layout) > 0:
    return False # means it wasn't removed
  if layout.tag in ("figure", "table", "formula"):
    return False
  return True

@dataclass
class _Patch:
  range: tuple[int, int]
  headline2chapter: dict[str, Chapter]
  layout_xmls_patches: dict[str, Element]

class _PatcherReader:
  def __init__(self, map_path: Path, no_matched_chapters: dict[int, Chapter]):
    self._pre_page_index: int = 0
    self._current_patch: _Patch | None = None
    self._xml_infos_iter: Iterator[XML_Info, None, None] = iter(xml_files(map_path))
    self._no_matched_chapters: dict[int, Chapter] = no_matched_chapters

  def read(self, page_index: int) -> _Patch | None:
    assert self._pre_page_index <= page_index
    self._pre_page_index = page_index

    while True:
      if self._current_patch is not None:
        patch_begin, patch_end = self._current_patch.range
        if page_index < patch_begin:
          return None
        if page_index <= patch_end:
          return self._current_patch

      xml_info = next(self._xml_infos_iter, None)
      if xml_info is None:
        self._current_patch = None
        return None

      file_path, _, begin, end = xml_info
      map_element = read_xml_file(file_path)
      self._current_patch = _Patch(
        range=(begin, end),
        headline2chapter={},
        layout_xmls_patches={},
      )
      for child in map_element:
        if child.tag != "mapper":
          continue
        headline_id = child.get("headline-id", None)
        chapter_id = int(child.get("chapter-id", "-1"))
        chapter = self._no_matched_chapters.pop(chapter_id, None)
        if headline_id is None or chapter is None:
          continue
        self._current_patch.headline2chapter[headline_id] = chapter

      patch_element = map_element.find("patch")
      if patch_element is not None:
        for child in patch_element:
          id = child.get("id", None)
          if id is None:
            continue
          self._current_patch.layout_xmls_patches[id] = child



import io
import re
import matplotlib.pyplot as plt

from xml.etree.ElementTree import fromstring, Element
from latex2mathml.converter import convert
from ..utils import sha256_hash
from .types import LaTeXRender
from .context import Context


def try_gen_table(context: Context, element: Element) -> list[Element] | None:
  if context.table_render == LaTeXRender.CLIPPING:
    return None

  table_html = _find_child(element, ("html",))
  children: list[Element] = []
  if table_html is not None:
    for child in table_html:
      children.append(child)

  return children

def try_gen_formula(context: Context, element: Element) -> Element | None:
  if context.latex_render == LaTeXRender.CLIPPING:
    return None

  latex = (element.text or "").strip()
  if not latex:
    return None

  latex_expr = _normalize_expression(latex)
  if context.latex_render == LaTeXRender.MATHML:
    return _latex2mathml(latex_expr)

  elif context.latex_render == LaTeXRender.SVG:
    svg_image = _latex_formula2svg(latex_expr)
    if svg_image is None:
      return None

    file_name = f"{sha256_hash(svg_image)}.svg"
    img_element = _create_image_element(file_name, element)
    context.add_asset(file_name, "image/svg+xml", svg_image)

    return img_element

def try_gen_asset(context: Context, element: Element) -> Element | None:
  hash = element.get("hash", None)
  if hash is None:
    return None

  file_name = f"{hash}.png"
  context.use_asset(file_name, "image/png")

  return _create_image_element(file_name, element)

_ESCAPE_UNICODE_PATTERN = re.compile(r"&#x([0-9A-Fa-f]{5});")

def _latex2mathml(latex: str) -> None | Element:
  try:
    html_latex = convert(latex)
  except Exception:
    return None

  # latex2mathml 转义会带上一个奇怪的 `&` 前缀，这显然是多余的
  # 不得已，在这里用正则表达式处理以修正这个错误
  def repl(match):
    hex_code = match.group(1)
    char = chr(int(hex_code, 16))
    if char == "<":
      return "&lt;"
    elif char == ">":
      return "&gt;"
    else:
      return char

  mathml = re.sub(
    pattern=_ESCAPE_UNICODE_PATTERN,
    repl=repl,
    string=html_latex,
  )
  try:
    return fromstring(mathml)
  except Exception:
    return None

def _latex_formula2svg(latex: str, font_size: int=12):
  # from https://www.cnblogs.com/qizhou/p/18170083
  try:
    output = io.BytesIO()
    plt.rc("text", usetex = True)
    plt.rc("font", size = font_size)
    fig, ax = plt.subplots()
    txt = ax.text(0.5, 0.5, f"${latex}$", ha="center", va="center", transform=ax.transAxes)
    ax.axis("off")
    fig.canvas.draw()
    bbox = txt.get_window_extent(renderer=fig.canvas.get_renderer())
    fig.set_size_inches(bbox.width / fig.dpi, bbox.height / fig.dpi)
    plt.savefig(
      output,
      format="svg",
      transparent=True,
      bbox_inches="tight",
      pad_inches=0,
    )
    return output.getvalue()
  except Exception:
    return None

def _create_image_element(file_name: str, origin: Element):
  img_element = Element("img")
  img_element.set("src", f"../assets/{file_name}")
  alt: str | None = None

  if origin.text:
    alt = origin.text
  if alt is None:
    img_element.set("alt", "image")
  else:
    img_element.set("alt", alt)

  return img_element

def _find_child(parent: Element, tags: tuple[str, ...]) -> Element | None:
  for child in parent:
    if child.tag in tags:
      return child
  return None

def _normalize_expression(expression: str) -> str:
  expression = expression.replace("\n", "")
  expression = expression.strip()
  return expression



import shutil

from pathlib import Path
from xml.etree.ElementTree import Element

from ..reference import samples, NumberStyle
from ..data import Paragraph, Layout, Line
from ..utils import Partition, PartitionTask
from .common import State, Corrector
from .paragraphs_reader import ParagraphsReader


class SingleCorrector(Corrector):
  def do(self, from_path: Path, request_path: Path, is_footnote: bool) -> None:
    request_path.mkdir(parents=True, exist_ok=True)
    partition: Partition[tuple[int, int], State, Element] = Partition(
      dimension=2,
      context=self.ctx,
      sequence=self.generate_request_xml(from_path),
      done=lambda _, __: self.ctx.reporter.increment(),
      remove=lambda begin, end: shutil.rmtree(
        request_path / _chunk_name(begin, end),
      ),
    )
    with partition:
      self.threads.run(
        next_task=partition.pop_task,
        thread_payload=lambda: ParagraphsReader(from_path),
        invoke=lambda reader, task: self._emit_request(
          task=task,
          reader=reader,
          request_path=request_path,
          is_footnote=is_footnote,
        ),
      )

  def _emit_request(
        self,
        task: PartitionTask[tuple[int, int], State, Element],
        reader: ParagraphsReader,
        request_path: Path,
        is_footnote: bool,
      ) -> Element:

    with task:
      begin = task.begin
      end = task.end
      request_element = task.payload
      chunk_element = self._correct_request(
        reader=reader,
        request_element=request_element,
        is_footnote=is_footnote,
      )
      self.ctx.write_xml_file(
        file_path=request_path / _chunk_name(begin, end),
        xml=chunk_element,
      )

  def _correct_request(self, reader: ParagraphsReader, request_element: Element, is_footnote: bool) -> Element:
    resp_element = self.llm.request_xml(
      template_name="correction/single",
      user_data=request_element,
      params={
        "layouts_count": 4,
        "is_footnote": is_footnote,
        "marks": samples(NumberStyle.CIRCLED_NUMBER, 6),
      },
    )
    raw_lines_list = list(self.extract_lines(request_element))
    corrected_lines_dict = dict(self.extract_lines(resp_element))
    paragraphs: dict[tuple[int, int], Paragraph] = {}

    for index, _ in sorted(raw_lines_list, key=lambda x: x[0]):
      page_index, order_index = index
      paragraph = reader.read(
        layout_index=(page_index, order_index),
      )
      if paragraph is None:
        continue

      para_index = (paragraph.page_index, paragraph.order_index)
      if para_index not in paragraphs:
        paragraphs[para_index] = paragraph

      corrected_lines = corrected_lines_dict.get(index, None)
      if corrected_lines is not None:
        self._apply_corrected_lines(
          paragraph=paragraph,
          index=index,
          corrected_lines=corrected_lines,
        )

    chunk_element = Element("chunk")

    for para_index in sorted(list(paragraphs.keys())):
      page_index, order_index = para_index
      paragraph = paragraphs[para_index]
      paragraph_element = paragraph.to_xml()
      paragraph_element.set("page-index", str(page_index))
      paragraph_element.set("order-index", str(order_index))
      chunk_element.append(paragraph_element)

    return chunk_element

  def _apply_corrected_lines(
        self,
        paragraph: Paragraph,
        index: tuple[int, int],
        corrected_lines: list[str],
      ) -> None:

    layout: Layout | None = None
    for got_layout in paragraph.layouts:
      if index == (got_layout.page_index, got_layout.order_index):
        layout = got_layout
    if layout is None:
      return None
    layout.lines = [
      Line(
        text=line_text,
        confidence="1.0",
      )
      for line_text in corrected_lines
    ]

def _chunk_name(begin: tuple[int, int], end: tuple[int, int]) -> str:
  return f"chunk_{begin[0]}_{begin[1]}_{end[0]}_{end[1]}.xml"


from __future__ import annotations
from json import loads
from dataclasses import dataclass
from pathlib import Path
from typing import Callable
from xml.etree.ElementTree import tostring, Element
from .i18n import I18N
from .template import Template


@dataclass
class NavPoint:
  index_id: int
  order: int
  file_name: str

def gen_index(
    template: Template,
    i18n: I18N,
    meta: dict,
    index_file_path: Path,
    has_cover: bool,
    check_chapter_exits: Callable[[int], bool],
  ) -> tuple[str, list[NavPoint]]:

  nav_elements: list[Element]
  nav_points: list[NavPoint]
  depth: int

  if index_file_path.exists():
    prefaces, chapters = _parse_index(index_file_path)
    nav_point_generation = _NavPointGeneration(
      has_cover=has_cover,
      check_chapter_exits=check_chapter_exits,
      chapters_count=(
        _count_chapters(prefaces) +
        _count_chapters(chapters)
      ),
    )
    nav_elements = []
    for chapters_list in (prefaces, chapters):
      for chapter in chapters_list:
        element = nav_point_generation.generate(chapter)
        if element is not None:
          nav_elements.append(element)

    depth = max(
      _max_depth(prefaces),
      _max_depth(chapters),
    )
    nav_points = nav_point_generation.nav_points

  else:
    nav_elements = []
    nav_points = []
    depth = 0

  toc_ncx = template.render(
    template="toc.ncx",
    depth=depth,
    i18n=i18n,
    meta=meta,
    has_cover=has_cover,
    nav_points=[tostring(p, encoding="unicode") for p in nav_elements],
  )
  return toc_ncx, nav_points

def _count_chapters(chapters: list[_Chapter]) -> int:
  count: int = 0
  for chapter in chapters:
    count += 1 + _count_chapters(chapter.children)
  return count

def _max_depth(chapters: list[_Chapter]) -> int:
  max_depth: int = 0
  for chapter in chapters:
    max_depth = max(
      max_depth,
      _max_depth(chapter.children) + 1,
    )
  return max_depth

class _NavPointGeneration:
  def __init__(self, has_cover: bool, chapters_count: int, check_chapter_exits: Callable[[int], bool]):
    self._nav_points: list[NavPoint] = []
    self._next_order: int = 2 if has_cover else 1
    self._digits = len(str(chapters_count))
    self._check_chapter_exits: Callable[[int], bool] = check_chapter_exits

  @property
  def nav_points(self) -> list[NavPoint]:
    return self._nav_points

  def generate(self, chapter: _Chapter) -> Element | None:
    if not self._check_chapter_exits(chapter.id):
      return None

    part_id = str(chapter.id).zfill(self._digits)
    file_name = f"part{part_id}.xhtml"
    order = self._next_order

    nav_point_xml = Element("navPoint")
    nav_point_xml.set("id", f"np_{chapter.id}")
    nav_point_xml.set("playOrder", str(order))

    label_xml = Element("navLabel")
    label_text_xml = Element("text")
    label_text_xml.text = chapter.headline
    label_xml.append(label_text_xml)

    content_xml = Element("content")
    content_xml.set("src", f"Text/{file_name}")

    nav_point_xml.append(label_xml)
    nav_point_xml.append(content_xml)

    self._next_order += 1
    self._nav_points.append(NavPoint(
      index_id=chapter.id,
      order=order,
      file_name=file_name,
    ))
    for child in chapter.children:
      child_xml = self.generate(child)
      if child_xml is not None:
        nav_point_xml.append(child_xml)

    return nav_point_xml

@dataclass
class _Chapter:
  id: int
  headline: str
  children: list[_Chapter]

def _parse_index(file_path: Path) -> tuple[list[_Chapter], list[_Chapter]]:
  data: dict | list
  with open(file_path, "r", encoding="utf-8") as file:
    data = loads(file.read())
  if isinstance(data, list):
    return [], _transform_chapters(data)
  elif isinstance(data, dict):
    return (
      _transform_chapters(data["prefaces"]),
      _transform_chapters(data["chapters"]),
    )

def _transform_chapters(data_list: list) -> list[_Chapter]:
  chapters: list[_Chapter] = []
  for data in data_list:
    chapters.append(_Chapter(
      id=int(data["id"]),
      headline=data["headline"],
      children=_transform_chapters(data["children"]),
    ))
  return chapters


import re
import shutil

from json import dumps
from pathlib import Path
from typing import Iterable
from xml.etree.ElementTree import Element

from ..xml import encode
from .contents import Contents
from .data import ASSET_LAYOUT_KINDS
from .utils import read_xml_file


_CHAPTER_FILE_PATTERN = re.compile(r"chapter(_\d+)?\.xml$")
_ASSET_FILE_PATTERN = re.compile(r"([0-9a-f]+)\.[a-zA-Z0-9]+$")

def output(
    contents: Contents | None,
    output_path: Path,
    meta_path: Path,
    chapter_output_path: Path,
    assets_path: Path,
  ) -> None:

  if contents is not None:
    index_path = output_path / "index.json"
    with open(index_path, "w", encoding="utf-8") as f:
      f.write(dumps(contents.json(), ensure_ascii=False, indent=2))

  cover_path = assets_path / "cover.png"
  output_chapters_path = output_path / "chapters"
  output_assets_path = output_path / "assets"

  if cover_path.exists():
    shutil.copy(cover_path, output_path / "cover.png")
  shutil.copy(meta_path, output_path / "meta.json")

  asset_hash_set: set[str] = set()
  output_chapters_path.mkdir(parents=True, exist_ok=True)

  for file in chapter_output_path.iterdir():
    if file.is_file() and _CHAPTER_FILE_PATTERN.match(file.name):
      chapter = _transform_chapter(file)
      asset_hash_set.update(_search_asset_hashes(chapter))
      target_path = output_chapters_path / file.name
      with open(target_path, "w", encoding="utf-8") as f:
        f.write(encode(chapter))

  if asset_hash_set:
    output_assets_path.mkdir(parents=True, exist_ok=True)
    for file in assets_path.iterdir():
      if not file.is_file():
        continue
      match = _ASSET_FILE_PATTERN.match(file.name)
      if match is None:
        continue
      asset_hash = match.group(1)
      if asset_hash not in asset_hash_set:
        continue
      shutil.copy(file, output_assets_path / file.name)

def _search_asset_hashes(chapter: Element):
  for chapter_child in chapter:
    if chapter_child.tag == "footnote":
      for footnote_child in chapter_child:
        hash = _get_asset_hash(footnote_child)
        if hash is not None:
          yield hash
    else:
      hash = _get_asset_hash(chapter_child)
      if hash is not None:
        yield hash

def _get_asset_hash(layout: Element) -> str | None:
  if layout.tag in ASSET_LAYOUT_KINDS:
    return layout.get("hash", None)
  else:
    return None

def _transform_chapter(origin_path: Path) -> Element:
  raw_chapter_element = read_xml_file(origin_path)
  chapter_element = Element(
    raw_chapter_element.tag,
    attrib=raw_chapter_element.attrib,
  )
  for child in raw_chapter_element:
    if child.tag == "footnote":
      child = _transform_footnote(child)
    else:
      child = _transform_layout(child)
    chapter_element.append(child)

  return chapter_element

def _transform_footnote(raw_footnote: Element) -> Element:
  footnote = Element(raw_footnote.tag)
  footnote.set("id", raw_footnote.get("id"))
  for child in raw_footnote:
    if child.tag == "mark":
      mark_element = Element(child.tag, attrib=child.attrib)
      footnote.append(mark_element)
      mark_element.text = (child.text or "").strip()
    else:
      layout = _transform_layout(child)
      footnote.append(layout)
  return footnote

def _transform_layout(raw_layout: Element) -> Element:
  layout = Element(raw_layout.tag)

  if raw_layout.tag not in ASSET_LAYOUT_KINDS:
    _fill_lines(layout, raw_layout)
  else:
    layout_hash = raw_layout.get("hash", None)
    if layout_hash is not None:
      layout.set("hash", layout_hash)
    for child in raw_layout:
      if child.tag == "caption":
        caption_element = Element(child.tag)
        layout.append(caption_element)
        _fill_lines(caption_element, child)
      elif child.text and child.tag == "latex":
        layout.text = child.text.strip()

  return layout

def _fill_lines(target_element: Element, line_elements: Iterable[Element]):
  text_buffer: list[str] = []
  last_mark_element: Element | None = None

  def flush_buffer():
    nonlocal text_buffer, last_mark_element
    if len(text_buffer) > 0:
      if last_mark_element is None:
        target_element.text = "".join(text_buffer)
      else:
        last_mark_element.tail = "".join(text_buffer)
      text_buffer.clear()

  for line_element in line_elements:
    if line_element.tag != "line":
      continue
    text = (line_element.text or "").strip()
    if text:
      text_buffer.append(text)
    for child in line_element:
      if child.tag == "mark":
        flush_buffer()
        target_element.append(child)
        last_mark_element = child
      text = (child.tail or "").strip()
      if text:
        text_buffer.append(text)

  flush_buffer()


import re

from doc_page_extractor import Rectangle
from shapely.geometry import Polygon


def rate(value1: float, value2: float) -> float:
  if value1 > value2:
    value1, value2 = value2, value1
  if value2 == 0.0:
    return 1.0
  else:
    return value1 / value2

def intersection_area_size(rect1: Rectangle, rect2: Rectangle) -> tuple[float, float]:
  poly1 = Polygon(rect1)
  poly2 = Polygon(rect2)
  intersection = poly1.intersection(poly2)

  if intersection.is_empty:
    return 0.0, 0.0

  if not isinstance(intersection, Polygon):
    return intersection.area, 0.0

  x1: float = float("inf")
  y1: float = float("inf")
  x2: float = float("-inf")
  y2: float = float("-inf")

  exterior = intersection.exterior
  if exterior is None:
    return 0.0, 0.0

  for x, y in exterior.coords:
    x1 = min(x1, x)
    y1 = min(y1, y)
    x2 = max(x2, x)
    y2 = max(y2, y)
  return x2 - x1, y2 - y1

_CJKA_PATTERN = re.compile(r"[\u4e00-\u9fff\u3040-\u309f\u30a0-\u30ff\uac00-\ud7a3\u0600-\u06ff]")

# 中、日、韩、阿拉伯文
def contains_cjka(text: str):
  return bool(_CJKA_PATTERN.search(text))


from typing import TypedDict
from strenum import StrEnum
from xml.etree.ElementTree import Element


class Phase(StrEnum):
  EXTRACTION = "extraction"
  TEXT_JOINT = "text-joint"
  FOOTNOTE_JOINT = "footnote-joint"
  COMPLETED = "completed"

class State(TypedDict):
  phase: Phase
  max_data_tokens: int
  max_paragraph_tokens: int
  max_paragraphs: int
  completed_ranges: list[list[int]]

class SequenceType(StrEnum):
  TEXT = "text"
  FOOTNOTE = "footnote"

class Truncation(StrEnum):
  YES = "truncated"
  NO = "not-truncated"
  PROBABLY = "probably"
  UNCERTAIN = "uncertain"

def get_truncation_attr(element: Element, attr_name: str) -> Truncation:
  value = element.get(attr_name, None)
  if value is not None:
    try:
      return Truncation(value)
    except Exception:
      pass
  return Truncation.UNCERTAIN


from .types import OCRLevel, PDFPageExtractorProgressReport
from .extractor import create_pdf_page_extractor, PDFPageExtractor
from .protocol import DocExtractorProtocol
from .types import (
  Block,
  AssetBlock,
  Text,
  TextKind,
  TextBlock,
  TableBlock,
  TableFormat,
  FormulaBlock,
  FigureBlock,
  ExtractedTableFormat,
)


from pathlib import Path

from ...llm import LLM
from ..reporter import Reporter, AnalysingStep
from ..utils import Context, MultiThreads
from .common import Phase, State, SequenceType
from .ocr_extractor import extract_ocr
from .joint import join


def extract_sequences(
      llm: LLM,
      reporter: Reporter,
      threads: MultiThreads,
      workspace_path: Path,
      ocr_path: Path,
      max_request_data_tokens: int,
      max_paragraph_tokens: int,
      max_paragraphs: int,
    ) -> None:

  context: Context[State] = Context(
    reporter=reporter,
    path=workspace_path,
    init=lambda: {
      "phase": Phase.EXTRACTION.value,
      "max_request_data_tokens": max_request_data_tokens,
      "max_paragraph_tokens": max_paragraph_tokens,
      "max_paragraphs": max_paragraphs,
      "completed_ranges": [],
    },
  )
  while context.state["phase"] != Phase.COMPLETED:
    if context.state["phase"] == Phase.EXTRACTION:
      reporter.go_to_step(AnalysingStep.EXTRACT_SEQUENCE)
      extract_ocr(
        llm=llm,
        context=context,
        threads=threads,
        ocr_path=ocr_path,
      )
      context.state = {
        **context.state,
        "phase": Phase.TEXT_JOINT.value,
        "completed_ranges": [],
      }
    elif context.state["phase"] == Phase.TEXT_JOINT:
      reporter.go_to_step(AnalysingStep.VERIFY_TEXT_PARAGRAPH)
      join(
        llm=llm,
        context=context,
        threads=threads,
        type=SequenceType.TEXT,
        extraction_path=workspace_path / Phase.EXTRACTION.value,
        join_path=workspace_path / Phase.TEXT_JOINT.value,
      )
      context.state = {
        **context.state,
        "phase": Phase.FOOTNOTE_JOINT.value,
        "completed_ranges": [],
      }
    elif context.state["phase"] == Phase.FOOTNOTE_JOINT:
      reporter.go_to_step(AnalysingStep.VERIFY_FOOTNOTE_PARAGRAPH)
      join(
        llm=llm,
        context=context,
        threads=threads,
        type=SequenceType.FOOTNOTE,
        extraction_path=workspace_path / Phase.EXTRACTION.value,
        join_path=workspace_path / Phase.FOOTNOTE_JOINT.value,
      )
      context.state = {
        **context.state,
        "phase": Phase.COMPLETED.value,
        "completed_ranges": [],
      }


from io import StringIO
from enum import auto, Enum
from typing import Generator
from dataclasses import dataclass


class TagKind(Enum):
  OPENING = auto()
  CLOSING = auto()
  SELF_CLOSING = auto()

@dataclass
class Tag:
  kind: TagKind
  name: str
  proto: str
  attributes: list[tuple[str, str]]

  def __str__(self):
    buffer = StringIO()
    buffer.write("<")
    if self.kind == TagKind.CLOSING:
      buffer.write("/")
    buffer.write(self.name)
    if len(self.attributes) > 0:
      buffer.write(" ")
      for i, (attr_name, attr_value) in enumerate(self.attributes):
        buffer.write(attr_name)
        buffer.write("=")
        buffer.write("\"")
        buffer.write(attr_value)
        buffer.write("\"")
        if i < len(self.attributes) - 1:
          buffer.write(" ")
    if self.kind == TagKind.SELF_CLOSING:
      buffer.write("/>")
    else:
      buffer.write(">")
    return buffer.getvalue()

  def find_invalid_name(self) -> str | None:
    for name in self._iter_tag_names():
      if not all(is_valid_value_char(c) for c in name):
        return name
      # https://www.w3schools.com/xml/xml_elements.asp
      # The following logic enforces a subset of XML naming rules:
      # - Names must not be empty.
      # - Names must start with a letter (a-z, A-Z) or an underscore (_).
      if name == "":
        return name
      char = name[0]
      if char == "_":
        continue
      if "a" <= char <= "z" or "A" <= char <= "Z":
        continue
      return name

    return None

  def find_invalid_attr_value(self) -> tuple[str, str] | None:
    for attr_name, attr_value in self.attributes:
      if not all(is_valid_value_char(c) for c in attr_value):
        return attr_name, attr_value
    return None

  def _iter_tag_names(self) -> Generator[str, None, None]:
    yield self.name
    for attr_name, _ in self.attributes:
      yield attr_name

def is_valid_value_char(char: str) -> bool:
  if is_valid_name_char(char):
    return True
  if char == ",":
    return True
  if char == ".":
    return True
  if char == "/":
    return True
  return False

def is_valid_name_char(char: str) -> bool:
  if "a" <= char <= "z":
    return True
  if "A" <= char <= "Z":
    return True
  if "0" <= char <= "9":
    return True
  if char == "_":
    return True
  if char == "-":
    return True
  return False


from typing import Any, Generator
from dataclasses import dataclass


@dataclass
class Contents:
  page_indexes: list[int]
  prefaces: list["Chapter"]
  chapters: list["Chapter"]

  def __iter__(self) -> Generator["Chapter", None, None]:
    for chapter in self.prefaces:
      yield chapter
      yield from chapter
    for chapter in self.chapters:
      yield chapter
      yield from chapter

  def json(self) -> dict[str, Any]:
    return {
      "prefaces": [chapter.json() for chapter in self.prefaces],
      "chapters": [chapter.json() for chapter in self.chapters],
    }

@dataclass
class Chapter:
  id: int
  name: str
  children: list["Chapter"]

  def __iter__(self) -> Generator["Chapter", None, None]:
    for child in self.children:
      yield child
      yield from child

  def json(self) -> dict[str, Any]:
    return {
      "id": self.id,
      "headline": self.name,
      "children": [child.json() for child in self.children],
    }


from pathlib import Path
from typing import Generator
from ..utils import xml_files


def list_chapter_files(chapter_path: Path) -> Generator[tuple[int | None, Path], None, None]:
  head_path = chapter_path.joinpath("chapter.xml")
  if head_path.exists():
    yield None, head_path

  for path, prefix, chapter_id, _ in xml_files(chapter_path):
    if prefix == "chapter":
      yield chapter_id, path


from typing import Generator
from xml.etree.ElementTree import tostring, Element

from .i18n import I18N
from .template import Template
from .context import Context
from .gen_asset import try_gen_table, try_gen_formula, try_gen_asset


def generate_part(
      context: Context,
      template: Template,
      chapter_xml: Element,
      i18n: I18N,
    ) -> str:

  return template.render(
    template="part.xhtml",
    i18n=i18n,
    content=[
      tostring(child, encoding="unicode")
      for child in _render_contents(context, chapter_xml)
    ],
    citations=[
      tostring(child, encoding="unicode")
      for child in _render_footnotes(context, chapter_xml)
    ],
  )

_XML2HTML_TAGS: dict[str, str] = {
  "headline": "h1",
  "quote": "p",
  "text": "p",
}

def _render_contents(context: Context, chapter_element: Element) -> Generator[Element, None, None]:
  for child in chapter_element:
    layout = _render_layout(context, child)
    if layout is not None:
      yield layout

def _render_footnotes(context: Context, chapter_element: Element):
  for footnote in chapter_element:
    if footnote.tag != "footnote":
      continue

    found_mark = False
    citation_div = Element("div", attrib={
      "class": "citation",
    })
    for child in footnote:
      if child.tag == "mark":
        found_mark = True
      else:
        layout = _render_layout(context, child)
        if layout is not None:
          citation_div.append(layout)

    if not found_mark or len(citation_div) == 0:
      continue

    footnote_id = int(footnote.get("id"))
    ref = Element("a")
    ref.text = f"[{footnote_id}]"
    ref.attrib = {
      "id": f"mark-{footnote_id}",
      "href": f"#ref-{footnote_id}",
      "class": "citation",
    }
    first_layout = citation_div[0]
    if first_layout.tag == "p":
      ref.tail = first_layout.text
      first_layout.text = None
      first_layout.insert(0, ref)
    else:
      inject_p = Element("p")
      inject_p.append(ref)
      citation_div.insert(0, inject_p)

    yield citation_div

def _render_layout(context: Context, raw_layout: Element) -> Element | None:
  if raw_layout.tag == "footnote":
    pass

  elif raw_layout.tag in _XML2HTML_TAGS:
    layout = Element(_XML2HTML_TAGS[raw_layout.tag])
    layout.text = raw_layout.text
    for mark in raw_layout:
      assert mark.tag == "mark"
      mark_id = int(mark.get("id"))
      anchor = Element("a")
      anchor.attrib = {
        "id": f"ref-{mark_id}",
        "href": f"#mark-{mark_id}",
        "class": "super",
      }
      layout.append(anchor)
      anchor.text = f"[{mark_id}]"
      anchor.tail = mark.tail

    if raw_layout.tag == "quote":
      blockquote = Element("blockquote")
      blockquote.append(layout)
      return blockquote
    else:
      return layout

  else:
    asset_wrapper = Element("div", attrib={
      "class": "alt-wrapper",
    })
    if raw_layout.tag == "table":
      asset_wrapper.extend(try_gen_table(context, raw_layout))
    elif raw_layout.tag == "formula":
      formula = try_gen_formula(context, raw_layout)
      if formula is not None:
        asset_wrapper.append(formula)

    if len(asset_wrapper) == 0:
      asset = try_gen_asset(context, raw_layout)
      if asset is not None:
        asset_wrapper.append(asset)

    if len(asset_wrapper) > 0:
      return asset_wrapper

  return None